	"                         V      ",
	"                         T W    ",
	"  Q T   KUSUPAT CCSTTT XXAPI    ",
	"HIUIRIBFISESILE HOTSTT CFLRN P  ",
	"UNILAOUPLRGRPRR LNOTIOIPSROC W  ",
	"PTTLPTSEL1V2EMM DTPPNUOUZMFH R  ",
